/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2005 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#if !defined(AFX_TRAY_H__4E26031B_8BE2_47C2_803A_FEBFB4262B31__INCLUDED_)
#define AFX_TRAY_H__4E26031B_8BE2_47C2_803A_FEBFB4262B31__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "TrayButton.h"
#include <algorithm>

class TrayTexture;
class TrayIcon;

//Alpha Transparency
#ifndef WS_EX_LAYERED
	#define WS_EX_LAYERED           0x00080000L
	#define LWA_COLORKEY            0x00000001L
	#define ULW_COLORKEY            0x00000001L
	#define LWA_ALPHA               0x00000002L
	#define ULW_ALPHA               0x00000002L
	#define ULW_OPAQUE              0x00000003L
#endif // WS_EX_LAYERED

#ifndef AC_SRC_ALPHA
	#define AC_SRC_ALPHA            0x00000001L
#endif

#ifndef WM_MOUSEWHEEL
	#define WM_MOUSEWHEEL 0x020A
#endif

#ifndef NIIF_USER
	#define NIIF_USER       0x00000004
#endif

typedef vector<TrayButton *> TrayButtonList;
typedef vector<TrayButton *>::iterator TrayButtonListIterator;

struct INFOTIPUPDATE_CONFIG
{
	int x;
	int y;
	TrayButton *button;
};

class Tray  
{

public:

	Tray(const string &name);
	virtual ~Tray();

	void load(HINSTANCE hInstance, HWND setbox = 0);
	void reconfigure(bool refresh = false);

	const HWND &gethWnd() const { return hWnd; }
	const string &getName() const { return name; }
	const HWND &getTooltip() const { return ToolTips; }
	const HWND &getBox() const { return box; }
	bool isAlwaysOnTop() const { return alwaysOnTop; }
	bool isVisible() const { return visible; }
	bool isStartHidden() const { return startHidden; }
	bool isLayeredFlagSet() const { return layeredFlag; }
	bool isAlphaMap() const { return alphaMap; }

	void reposition(int type, int newx, int newy, int newwidth, int newheight, int steps = 0, int time = 0, bool repositionBy = false);

	void setAlwaysOnTop(bool setAlwaysOnTop);
	void setAlpha(int newAlphaTransparency);
	void setAlphaFade(int newAlphaTransparency, int step, int delay);

	void repaint();
	void updateRegion();

	void show();	
	void hide();

	void ToggleHiddenIcons();

	void CreateLayout(bool clearBackground = true, int width = -1, int height = -1);

	void ScrollIcons(bool next, int amount);

	//For ReposButtons
	TrayButton *movecapture;
	//For Tooltips
	TrayButton *capture;
	
	bool showToolTip;

	bool transparencyMode;

	TrayTexture *traytexture;
	TrayIcon *trayicon;

	vector<CLASS_CONFIG> HideList;
	vector<CLASS_CONFIG> NoDragList;
	vector<CLASS_CONFIG> SortList;
	
	bool hideIfEmpty;
	int backupExportedButtonCount;

private:

	bool blocked;

	bool ParentDestroyNotificationReceived;
	
	TrayButton *dontPaintAddedTrayIcon;

	int internalAddCount;

	INFOTIPUPDATE_CONFIG UpdateConfig;

	HBITMAP oldbackupBMP;
	HDC backupDC;
	HBITMAP backupBitmap;
	bool skipBackup;

	bool autoSize;
	int autoSizeSteps;
	int autoSizeDelay;

	int startWithPosition;

	int cleanUpInterval;

	bool nodragndrop;

	string afterClass;
	string afterTitle;

	string name;

	bool alphaMap;
	bool alphaFade;
	int UpdateAlpha;
	int alphaFadeStep;
	int alphaFadeDelay;
	int alphaFadeBangStep;
	int alphaFadeBangEndValue;

	bool alphaInitialized;
	bool layeredFlag;
	int alphaTransparency;

	HWND hWnd;
	HINSTANCE hInstance;
	HWND box;
	HWND ToolTips;

	bool mouseInside;
	
	bool focusOnEnter;
	bool skipButton;

	TrayButtonList traybuttons;

	bool inCriticalProcess;

	int maxIconSize;

	bool autoHidden;

	int actualx;
	int actualy;
	int actualwidth;
	int actualheight;

	int leftBorder;
	int topBorder;
	int rightBorder;
	int bottomBorder;

	int layout;
	int direction;
	int wrapDirection;

	int lines_rows;
	int lines_rowsAdded;
	int wrapCount;
	int maxlines_rows;

	bool fireClicksAlways;

	bool startHidden;
	bool disablezOrderChange;
	bool alwaysOnTop;
	bool visible;

	int visibleIcons;
	TrayButton *changeState;
	bool hiddenvisible;

	string trayLButtonDown;
	string trayLButtonUp;
	string trayRButtonDown;
	string trayRButtonUp;
	string trayMButtonDown;
	string trayMButtonUp;
	string trayMWheelDown;
	string trayMWheelUp;
	string trayEnter;
	string trayLeave;
	string trayAdd;
	string trayRemove;
	string trayMove;
	string trayResize;

	string trayLineAdd;
	string trayLineRemove;

	bool borderDrag;

	//General Button Settings
	int xSpacing;
	int ySpacing;

	bool autoSizeFunction();

	bool IconAdd(NOTIFYICONDATA *pnid);
	bool IconModify(NOTIFYICONDATA *pnid);
	int IconDelete(NOTIFYICONDATA *pnid);

	bool IconCleanUp();

	void SortTrayIcons(vector<TrayButton *> &traybuttons)
	{
		std::sort( traybuttons.begin(), traybuttons.end(), CompareTrayIcons );
	}

	static bool CompareTrayIcons(TrayButton *b1, TrayButton *b2)
	{
		if ( b1->userHidden == b2->userHidden )
		{
			if ( b1->fixedPositionInTray < b2->fixedPositionInTray )
				return true;
			else
				return false;
		}
		else
		{
			if ( !b1->userHidden && b2->userHidden )
				return true;
			else
				return false;
		}
	};

protected:

	virtual TrayButton* onLButtonDblClk(int x, int y);
	virtual TrayButton* onLButtonDown(int x, int y, LPARAM lParam);
	virtual TrayButton* onLButtonUp(int x, int y);
	virtual TrayButton* onMButtonDblClk(int x, int y);
	virtual TrayButton* onMButtonDown(int x, int y);
	virtual TrayButton* onMButtonUp(int x, int y);
	virtual TrayButton* onRButtonDblClk(int x, int y);
	virtual TrayButton* onRButtonDown(int x, int y);
	virtual TrayButton* onRButtonUp(int x, int y); 
	virtual void onWheelDown(int x, int y);
	virtual void onWheelUp(int x, int y);
	virtual void onMouseEnter();
	virtual void onMouseLeave();
	virtual TrayButton* onMouseMove(int x, int y);
	virtual void onPaint(HDC hDC);
	virtual void onWindowPosChanging(WINDOWPOS *windowPos);
	virtual void onMove(int x, int y);
	virtual void onSize(int width, int height);
	virtual void onTimer(int timerID);

private:

	void updateHint(HWND hWnd, LPSTR caption, bool updatesize, bool remove = false);

private:

	void UpdateWindowTransparent(int newAlphaTransparency = 255);

	TrayButton *ButtonAtPoint( int x, int y );

private:

	BOOL onWindowMessage( UINT message, WPARAM wParam, LPARAM lParam, LRESULT &lResult );

public:

	static LRESULT WINAPI windowProcedure( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam );

	HWND InfoTiphWnd;

	bool infoTipsEnabled;

	bool showInfoTips;
	bool CreateInfoTipWindow();
	void ShowInfoTip(TrayButton *button, int x, int y);
	void HideInfoTip();
	void DestroyInfoTip();

	COLORREF InfoTipColor;
	COLORREF InfoTipTextColor;
};

#endif // !defined(AFX_TRAY_H__4E26031B_8BE2_47C2_803A_FEBFB4262B31__INCLUDED_)
