/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2005 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "stdafx.h"
#include "TrayButton.h"
#include "Tray.h"
#include "TrayIcon.h"

typedef BOOL (WINAPI *FUNC_BOOL__DWORD)(DWORD);
#ifndef ASFW_ANY
	#define ASFW_ANY ((DWORD)-1)
#endif

static FUNC_BOOL__DWORD pAllowSetForegroundWindow = NULL;

#define INVALID_ICON(x) ((NULL == (x)) || (1 == (int)(x)) || (-1 == (int)(x)))

#define INFOTIP_MAXIMUM_TIMEOUT 15000
#define INFOTIP_MINIMUM_TIMEOUT 7500

TrayButton::TrayButton( Tray *actualtray, HWND sethOwnerWnd, UINT setuID ):buttonx(0), buttony(0), buttonheight(0), buttonwidth(0)
{
	tray = actualtray;

	hOwnerWnd = sethOwnerWnd;
	uID = setuID;

	uCallbackMessage = 0;

	strTooltip = "";

	strTitle = "";
	strInfo = "";
	iTimeOut = 10000;
	dwInfoFlags = NIIF_NONE;
	
	hIcon = NULL;

	xTrayHideChecked = false;
	userHidden = false;
	useHover = false;
	noDragnDrop = -1;

	fixedPositionInTray = 100;
	positionInTray = -1;

	hoverDelta = (tray->trayicon->getIconSize() - tray->trayicon->getHoverIconSize()) / 2;
	hoverSize = tray->trayicon->getHoverIconSize();

	pAllowSetForegroundWindow = (FUNC_BOOL__DWORD)GetProcAddress( GetModuleHandle(_T("USER32")), "AllowSetForegroundWindow");
}

TrayButton::~TrayButton()
{
	if (hIcon)
		DestroyIcon(hIcon);
}

bool TrayButton::HasPoint( int xP, int yP )
{
	if ( xP < (buttonx + buttonwidth) && xP >= buttonx )
	{
		if ( yP < (buttony + buttonheight) && yP >= buttony )
			return true;
	}

	return false;
}

void TrayButton::onButtonClick(UINT nMessage, WPARAM wParam, LPARAM lParam)
{
	if ( uCallbackMessage == 0 )
		return;

	if ( nMessage != WM_MOUSEMOVE && pAllowSetForegroundWindow != NULL )
		pAllowSetForegroundWindow(ASFW_ANY);

	PostMessage( hOwnerWnd, uCallbackMessage, (WPARAM)uID, (LPARAM)nMessage	);
}

void TrayButton::setButtonLayout( int x, int y, int width, int height )
{
	buttonx = x;
	buttony = y;
	buttonwidth = width;	
	buttonheight = height;
}

void TrayButton::setIcon( HICON sethIcon )
{ 	
	if (hIcon)
	{
		DestroyIcon(hIcon);
		hIcon = NULL;
	}

	hIcon = CopyIcon( sethIcon );

	if ( hIcon == NULL )
		hIcon = sethIcon;
}

bool TrayButton::setTooltip( string setstrTooltip )
{
	if ( stricmp(strTooltip.c_str(), setstrTooltip.c_str()) != NULL )
	{
		strTooltip = setstrTooltip;
		return true;
	}
	else
	{
		return false;
	}
}

bool TrayButton::Update(NOTIFYICONDATA *pnid, bool &changeActiveTooltip, bool &autoShowInfoTip)
{
	bool bUpdated = false;

	uID = pnid->uID;

	if ( (pnid->uFlags & NIF_MESSAGE) != 0 )
		uCallbackMessage = pnid->uCallbackMessage;

	if ( (pnid->uFlags & NIF_TIP) != 0 )
	{
		if ( setTooltip( (string)pnid->szTip ) && useHover )
			changeActiveTooltip = true;
	}

	if ( (pnid->uFlags & NIF_ICON) != 0 && pnid->hIcon != NULL )
	{
		setIcon( pnid->hIcon );
		bUpdated = true;
	}

	if ( (pnid->uFlags & NIF_INFO) != 0 && IsOS( OS_2KXP ) )
	{
		if ( (int) pnid->uTimeout >= 0 )
			iTimeOut = min( max((int) pnid->uTimeout, INFOTIP_MINIMUM_TIMEOUT), INFOTIP_MAXIMUM_TIMEOUT);

		if ( stricmp( strTitle.c_str(), pnid->szInfoTitle) != NULL || stricmp( strInfo.c_str(), pnid->szInfo) != NULL )
			autoShowInfoTip = true;

		strInfo = (string) pnid->szInfo;
		strTitle = (string) pnid->szInfoTitle;

		dwInfoFlags = pnid->dwInfoFlags;		
	}

	return bUpdated;
}

void TrayButton::onPaint( HDC destDC )
{
	if ( !useHover )
		tray->trayicon->apply(destDC, buttonx, buttony, buttonwidth, buttonheight, 0, hIcon, tray->isAlphaMap());
	else
		tray->trayicon->apply(destDC, buttonx+hoverDelta, buttony+hoverDelta, hoverSize, hoverSize, 1, hIcon, tray->isAlphaMap());
}
