/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2005 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "stdafx.h"
#include "TrayIcon.h"

TrayIcon::TrayIcon(const string &name)
{
	defaulticon.iconSize = GetxRCInt(name, "IconSize", 16, 8, 64);

	defaulticon.icon_saturation = GetxRCInt(name, "IconSaturationIntensity", 255, 0, 255);
	defaulticon.icon_hueintensity = GetxRCInt(name, "IconHueIntensity", 0, 0, 255);
	defaulticon.icon_huecolor = GetxRCColor(name, "IconHueColor", RGB(255, 255, 255));
}

TrayIcon::~TrayIcon()
{
}

void TrayIcon::subconfigure(string type, ICONSETTINGS &workSettings, ICONSETTINGS &defaultSettings)
{
	workSettings.iconSize = GetxRCInt(type, "IconSize", defaultSettings.iconSize, 8, 64);
	
	workSettings.icon_saturation = GetxRCInt(type, "IconSaturationIntensity", defaultSettings.icon_saturation, 0, 255);
	workSettings.icon_hueintensity = GetxRCInt(type, "IconHueIntensity", defaultSettings.icon_hueintensity, 0, 255);
	workSettings.icon_huecolor = GetxRCColor(type, "IconHueColor", defaultSettings.icon_huecolor);
}

void TrayIcon::configure(const string &name)
{
	subconfigure((name + "Hover"), hovericon, defaulticon);
}

void TrayIcon::apply(HDC hDC, int x, int y, int width, int height, int buttonType, HICON hIcon, bool useAlphaMap)
{
	if(hIcon == NULL)
		return;

	int iconX = 0;
	int iconY = 0;
	int iconSize = 16;

	int _saturation = 255;
	int _hueIntensity = 0;
	COLORREF _hueColor = RGB(255, 255, 255);

	if (buttonType == 0)
	{
		iconSize = defaulticon.iconSize;
		_saturation = defaulticon.icon_saturation;
		_hueIntensity = defaulticon.icon_hueintensity;
		_hueColor = defaulticon.icon_huecolor;
	}
	else if (buttonType == 1)
	{
		iconSize = hovericon.iconSize;
		_saturation = hovericon.icon_saturation;
		_hueIntensity = hovericon.icon_hueintensity;
		_hueColor = hovericon.icon_huecolor;
	}

	if ( useAlphaMap || IsOS(OS_XP) || _saturation != 255 || _hueIntensity != 0 )
	{
		BITMAPINFO alpha;
		ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
		alpha.bmiHeader.biWidth=iconSize;      // Set size you need
		alpha.bmiHeader.biHeight=iconSize;    // Set size you need
		alpha.bmiHeader.biPlanes=1;
		alpha.bmiHeader.biBitCount=32;            // Can be 8, 16, 32 bpp or other number
		alpha.bmiHeader.biSizeImage=0;
		alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
		alpha.bmiHeader.biClrUsed= 0;
		alpha.bmiHeader.biClrImportant= 0;
		VOID *pvBits;

		HDC hdcBGCOPY = CreateCompatibleDC(hDC);
		HBITMAP	hbmBGCOPY = CreateDIBSection( hdcBGCOPY,
										&alpha,
										DIB_RGB_COLORS,
										&pvBits,
										NULL,
										0 );
		HBITMAP oldBGCOPYBMP = (HBITMAP) SelectObject(hdcBGCOPY, hbmBGCOPY);

		//Copy RealBG in Result Background
		BitBlt( hdcBGCOPY, 0, 0, iconSize, iconSize, hDC, x+iconX, y+iconY, SRCCOPY );

		HDC hdcBuffer = CreateCompatibleDC(hDC);
		HBITMAP	hbmBuffer = CreateDIBSection( hdcBuffer,
										&alpha,
										DIB_RGB_COLORS,
										&pvBits,
										NULL,
										0 );
		HBITMAP oldBufferBMP = (HBITMAP) SelectObject(hdcBuffer, hbmBuffer);

		//Copy RealBG in Result Background
		if ( IsOS(OS_2KXP) )
			BitBlt(hdcBuffer, 0, 0, iconSize, iconSize, hDC, x+iconX, y+iconY, SRCCOPY );

		//Draw Icon
		DrawIconEx(hdcBuffer, 0, 0, hIcon, iconSize, iconSize, 0, NULL, DI_NORMAL );

		HDC hdcMask = CreateCompatibleDC(hdcBuffer);
		HBITMAP	hbmMask = CreateDIBSection( hdcMask,
										&alpha,
										DIB_RGB_COLORS,
										&pvBits,
										NULL,
										0 );
		HBITMAP oldMaskBMP = (HBITMAP) SelectObject(hdcMask, hbmMask);

		//Draw Mask of the Icon
		DrawIconEx(hdcMask, 0, 0, hIcon, iconSize, iconSize, 0, NULL, DI_MASK);

		if ( IsOS(OS_XP) )
		{
			BITMAP bm;
			GetObject(hbmBuffer, sizeof(BITMAP), &bm);

			BITMAP bmmask;
			GetObject(hbmMask, sizeof(BITMAP), &bmmask);

			BITMAP bmorg;
			GetObject(hbmBGCOPY, sizeof(BITMAP), &bmorg);

			BYTE *pPixel = (LPBYTE)bm.bmBits;
			BYTE *pPixelMask = (LPBYTE)bmmask.bmBits;
			BYTE *pPixelOrg = (LPBYTE)bmorg.bmBits;

			BYTE r = 0;
			BYTE g = 0;
			BYTE b = 0;
			bool skip = false;

			for (int i = 0; i < bm.bmHeight * bm.bmWidth; ++i)
			{
				//Filter Mask and insert OriginalBG to preserve AlphaChannel
				if ( pPixelMask[0] == 255 && pPixelMask[1] == 255 && pPixelMask[2] == 255 )
				{
					pPixel[0] = pPixelOrg[0];	
					pPixel[1] = pPixelOrg[1];	
					pPixel[2] = pPixelOrg[2];	
					pPixel[3] = pPixelOrg[3];	

					skip = true;
				}
				//For colorizing without AlphaMap
				if ( !skip && pPixelOrg[0] == 255 && pPixelOrg[1] == 0 && pPixelOrg[2] == 255 && pPixel[3] != 255 && pPixel[3] != 0 )
				{
					pPixel[0] = pPixelOrg[0];	
					pPixel[1] = pPixelOrg[1];	
					pPixel[2] = pPixelOrg[2];	
					pPixel[3] = pPixelOrg[3];	

					skip = true;
				}

				if ( skip )
				{
					pPixel+= 4;
					pPixelMask+= 4;
					pPixelOrg+= 4;

					skip = false;
					continue;
				}

				r = pPixel[0];
				g = pPixel[1];
				b = pPixel[2];

				//AlphaMap Colorizing
				//Saturation Effect
				if( _saturation != 255 )
				{
					BYTE gray = (BYTE)(r * 0.3086 + g * 0.6094 + b * 0.0820);
					r = (BYTE)((r * _saturation + gray * (255 - _saturation) + 255) >> 8);
					g = (BYTE)((g * _saturation + gray * (255 - _saturation) + 255) >> 8);
					b = (BYTE)((b * _saturation + gray * (255 - _saturation) + 255) >> 8);
				}
				// Hue effect
				if( _hueIntensity != 0 )
				{
					// The B & R values of the hue is swapped here, can somebody tell me why it only works that way?
					// test normal
					r = (BYTE)((r * (255 - _hueIntensity) + GetBValue(_hueColor) * _hueIntensity ) >> 8);
					g = (BYTE)((g * (255 - _hueIntensity) + GetGValue(_hueColor) * _hueIntensity ) >> 8);
					b = (BYTE)((b * (255 - _hueIntensity) + GetRValue(_hueColor) * _hueIntensity ) >> 8);
				}

				pPixel[0] = r;
				pPixel[1] = g;
				pPixel[2] = b;
				//Fix for Icons without AlphaChannel (all is 0!), with AlphaChannel 0 is Mask
				if ( pPixel[3] == 0 )
					pPixel[3] = 255;

				pPixel+= 4;
				pPixelMask+= 4;
				pPixelOrg+= 4;
			}

			BitBlt( hDC, x+iconX, y+iconY, iconSize, iconSize, hdcBuffer, 0, 0, SRCCOPY);
		}
		else
		{
			int iy = 0;
			int ix = 0;
			COLORREF cl = 0;
			BYTE r = 0;
			BYTE g = 0;
			BYTE b = 0;
			for(iy = 0; iy < iconSize; iy++)
			{
				for(ix = 0; ix < iconSize; ix++)
				{
					// Loop through all transparent pixels...
					// If the mask-pixel is white or grey(if that is possible?), then break
					if( !GetPixel(hdcMask, ix, iy) )
					{
						cl = GetPixel(hdcBuffer, ix, iy);
						r = GetRValue(cl);
						g = GetGValue(cl);
						b = GetBValue(cl);

						// Saturation effect
						if( _saturation != 255 )
						{
							BYTE gray = (BYTE)(r * 0.3086 + g * 0.6094 + b * 0.0820);

							r = (BYTE)((r * _saturation + gray * (255 - _saturation) + 255) >> 8);
							g = (BYTE)((g * _saturation + gray * (255 - _saturation) + 255) >> 8);
							b = (BYTE)((b * _saturation + gray * (255 - _saturation) + 255) >> 8);
						}

						// Hue effect
						if( _hueIntensity != 0 )
						{
							r = (BYTE)((r * (255 - _hueIntensity) + GetRValue(_hueColor) * _hueIntensity + 255) >> 8);
							g = (BYTE)((g * (255 - _hueIntensity) + GetGValue(_hueColor) * _hueIntensity + 255) >> 8);
							b = (BYTE)((b * (255 - _hueIntensity) + GetBValue(_hueColor) * _hueIntensity + 255) >> 8);
						}

						SetPixelV(hdcBGCOPY, ix, iy, RGB(r,g,b));
					}
				}
			}

			BitBlt( hDC, x+iconX, y+iconY, iconSize, iconSize, hdcBGCOPY, 0, 0, SRCCOPY);
		}
				
		hbmMask = (HBITMAP) SelectObject(hdcMask, oldMaskBMP);
		DeleteObject(hbmMask);
		DeleteDC(hdcMask);

		hbmBuffer = (HBITMAP) SelectObject(hdcBuffer, oldBufferBMP);
		DeleteObject(hbmBuffer);
		DeleteDC(hdcBuffer);

		hbmBGCOPY = (HBITMAP) SelectObject(hdcBGCOPY, oldBGCOPYBMP);
		DeleteObject(hbmBGCOPY);
		DeleteDC(hdcBGCOPY);
	}
	else
	{
		DrawIconEx(hDC, x+iconX, y+iconY, hIcon, iconSize, iconSize, 0, NULL, DI_NORMAL);
	}
}
