/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2005 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "stdafx.h"
#include "TrayTexture.h"

TrayTexture::TrayTexture()
{
	transparencyNeeded = false;

	useAlphaMap = false;

	Gradient.clear();

	memoryDC = 0;
	memoryBitmap = 0;

	useSolidColors = false;
}

TrayTexture::~TrayTexture()
{
	Gradient.clear();

	if(memoryBitmap)
	{
		memoryBitmap = (HBITMAP) SelectObject(memoryDC, memoryBitmap);
		DeleteObject(memoryBitmap);
	}
	if(memoryDC)
		DeleteDC(memoryDC);
}

void TrayTexture::parseCFG(string cfg, COLORREF &bgcolor, COLORREF &lightcolor, COLORREF &darkcolor)
{
	while ( strstr( cfg.c_str(), "#" ) != NULL )
		cfg.replace(cfg.find("#"), 1, "", 0);

	int color;
	char main[32], light[32], dark[32];
	char *buffers[] = {main, light, dark};

	int numTokens = LCTokenize(cfg.c_str(), buffers, 3, NULL);

	if(numTokens == 1)
	{
		color = strtol(main, NULL, 16);
		bgcolor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
		lightcolor = GetSysColor(COLOR_3DLIGHT);
		darkcolor = GetSysColor(COLOR_3DSHADOW);
	}
	else if (numTokens == 2)
	{
		color = strtol(main, NULL, 16);
		bgcolor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));

		color = strtol(light, NULL, 16);
		lightcolor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
		darkcolor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
	}
	else if (numTokens == 3)
	{
		color = strtol(main, NULL, 16);
		bgcolor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));

		color = strtol(light, NULL, 16);
		lightcolor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));

		color = strtol(dark, NULL, 16);
		darkcolor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
	}
}

void TrayTexture::parseBevelCFG(string cfg, int &leftBevel, int &rightBevel, int &topBevel, int &bottomBevel)
{
	char left[32], right[32], top[32], bottom[32];
	char *buffers[] = {left, right, top, bottom };

	leftBevel = 0;
	rightBevel = 0;
	topBevel = 0;
	bottomBevel = 0;

	int numTokens = LCTokenize(cfg.c_str(), buffers, 4, NULL);
		
	if(numTokens == 1)
	{
		leftBevel = atoi(left);
		rightBevel = leftBevel;
		topBevel = leftBevel;
		bottomBevel = leftBevel;
	}
	else if (numTokens == 2)
	{
		leftBevel = atoi(left);
		rightBevel = leftBevel;
		topBevel = atoi(right);
		bottomBevel = topBevel;
	}
	else if (numTokens == 4)
	{
		leftBevel = atoi(left);
		rightBevel = atoi(right);
		topBevel = atoi(top);
		bottomBevel = atoi(bottom);
	}
}

void TrayTexture::configure(const string &name, const bool &transparencyMode)
{
	NameValuePair modeConstants[] =
	{
		{ "stretch", MULTIBLT_STRETCH },
		{ "tile", MULTIBLT_TILE },
		{ "tile-horizontal", MULTIBLT_TILEHORIZONTAL },
		{ "tile-vertical", MULTIBLT_TILEVERTICAL },
		{ "tilehorizontal", MULTIBLT_TILEHORIZONTAL },
		{ "tilevertical", MULTIBLT_TILEVERTICAL },
		{ 0, 0 }
	};

	NameValuePair borderValues[] = 
	{
		{ ".none", 0 },
		{ "none", 0 },
		{ "raised", 1 },
		{ "sunken", 2 },
		{ "etched", 3 },
		{ "bump", 4 },
		{ 0, 0 }
	};

	if ( IsOS(OS_2KXP) && !GetxRCBool(name, "UseSolidColors") && !GetxRCString(name, "Image", "").empty() )
		useAlphaMap = GetxRCBool(name, "AlphaMap");

	//Is useable in SolidColors and Image Mode
	borderMethod = GetxRCNamedValue(name, "BorderMethod", borderValues, 0);

	useSolidColors = ( !(GetxRCLine(name, "SolidColors", "")).empty() && (GetxRCString(name, "Image", "")).empty() );
	
	string imageFile = "";

	if ( useSolidColors )
	{
		imageFile = GetxRCLine(name, "SolidColors", "");
		
		parseCFG(imageFile, backgroundColor, lightBevelColor, darkBevelColor);
	
		//BG Insert in Gradient List
		Gradient.insert(Gradient.end(), backgroundColor );

		if (backgroundColor == RGB(255,0,255) )
			transparencyNeeded = true; 

		imageFile = GetxRCLine(name, "SolidBevelSize", "0");
		parseBevelCFG(imageFile, bevelSizeLeft, bevelSizeRight, bevelSizeTop, bevelSizeBottom);

		verticalGradient = GetxRCBool(name, "SolidGradientVertical");

		repeatedGradient = GetxRCInt(name, "SolidGradientRepeated", 0, 0);

		imageFile = GetxRCLine(name, "SolidGradientColors", "");
	
		if ( !imageFile.empty() )
		{	
			while ( strstr( imageFile.c_str(), "#" ) != NULL )
				imageFile.replace(imageFile.find("#"), 1, "", 0);

			int color;
			COLORREF insertColor;
			char token[4096];
			LPCSTR nextToken = imageFile.c_str();

			while (GetToken(nextToken, token, &nextToken, false))
			{
				color = strtol(token, NULL, 16);
				insertColor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
		
				Gradient.insert(Gradient.end(), insertColor );
			}
		}
	}
	else
	{
		imageFile = GetxRCString(name, "Image", "");

		memoryBitmap = LoadLSImage(imageFile.c_str(), 0);

		if (memoryBitmap)
		{
			GetLSBitmapSize(memoryBitmap, &imageWidth, &imageHeight);

			memoryDC = CreateCompatibleDC(NULL);
			SelectObject(memoryDC, memoryBitmap);

			taskbar_saturation = GetxRCInt(name , "SaturationIntensity", 255, 0, 255);
			taskbar_hueIntensity = GetxRCInt(name , "HueIntensity", 0, 0, 255);
			taskbar_hueColor = GetxRCColor(name, "HueColor", RGB(0, 0, 0));

			leftEdge = GetxRCInt(name, "ImageLeftEdge", 0, 0);
			topEdge = GetxRCInt(name, "ImageTopEdge", 0, 0);
			rightEdge = GetxRCInt(name, "ImageRightEdge", 0, 0);
			bottomEdge = GetxRCInt(name, "ImageBottomEdge", 0, 0);

			mode = GetxRCNamedValue(name, "ImageMode", modeConstants, MULTIBLT_STRETCH);

			if (useAlphaMap)
			{
				BITMAP bm;
				GetObject(memoryBitmap, sizeof(BITMAP), &bm);

				if ((bm.bmBits != 0) && (bm.bmBitsPixel == 32))
					PrepareAlphaChannel(memoryBitmap);
			}
			else
			{
				paintEffects(memoryDC, imageWidth, imageHeight);
			}
		}
		else
		{
			transparencyNeeded = true;
		}
	}
}

void TrayTexture::apply(HDC hDC, int x, int y, int width, int height, bool trueTransparency)
{
	if (useSolidColors)
	{
		RECT r;
		r.left = x;
		r.top = y;
		r.right = r.left + width;
		r.bottom = r.top + height;

		int bevelWidth = bevelSizeLeft+bevelSizeRight;
		int bevelHeight = bevelSizeTop+bevelSizeBottom;

		// fill in the background
		if (bevelWidth+bevelHeight > 0 && (bevelWidth < width && bevelHeight < height))
		{
			::SetBkColor(hDC, darkBevelColor);
			::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);

			r.right = r.right-bevelSizeRight;
			r.bottom = r.bottom-bevelSizeBottom;
			::SetBkColor(hDC, lightBevelColor);
			::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);

			r.left = r.left+bevelSizeLeft;
			r.top = r.top+bevelSizeTop;

			if ( Gradient.size() > 1 )
			{
				solidGradientPaint(hDC, r);
			}
			else
			{
				if (!trueTransparency && backgroundColor == RGB(255,0,255))
				{
					PaintDesktopEx(hDC, r.left, r.top, r.right-r.left, r.bottom-r.top, x, y, FALSE);
				}
				else
				{
					::SetBkColor(hDC, backgroundColor);
					::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);
				}
			}
		}
		else
		{
			if ( Gradient.size() > 1 )
			{
				solidGradientPaint(hDC, r);
			}
			else
			{
				if ( backgroundColor != RGB(255,0,255) )
				{
					::SetBkColor(hDC, backgroundColor);
					::ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);
				}
			}
		}
	}
	else
	{
		if (memoryBitmap == 0)
			return;

		if (trueTransparency)
		{
			SetStretchBltMode(hDC, STRETCH_DELETESCANS);

			MultiBlt(hDC, x, y, width, height,
						memoryDC, 0, 0, imageWidth, imageHeight,
						leftEdge, topEdge, rightEdge, bottomEdge,
						mode, SRCCOPY);
		}
		else if (!trueTransparency)
		{
			HDC bufferDC = CreateCompatibleDC(hDC);
			HBITMAP bufferBitmap = CreateCompatibleBitmap(hDC, width, height);
			HBITMAP oldBMP = (HBITMAP) SelectObject(bufferDC, bufferBitmap);

			SetStretchBltMode(bufferDC, STRETCH_DELETESCANS);

			MultiBlt(bufferDC, 0, 0, width, height,
					memoryDC, 0, 0, imageWidth, imageHeight,
					leftEdge, topEdge, rightEdge, bottomEdge,
					mode, SRCCOPY);
				
			TransparentBltLS(hDC, x, y, width, height,
					bufferDC, 0, 0,
					RGB(255, 0, 255));

			bufferBitmap = (HBITMAP) SelectObject(bufferDC, oldBMP);
			DeleteObject(bufferBitmap);
			DeleteDC(bufferDC);
		}
	}

	RECT completeFillRect = {x, y, x+width, y+height};
	if (borderMethod == 1)
		DrawEdge(hDC, &completeFillRect, EDGE_RAISED, BF_RECT | BF_SOFT);
	else if (borderMethod == 2)
		DrawEdge(hDC, &completeFillRect, EDGE_SUNKEN, BF_RECT | BF_SOFT);
	else if (borderMethod == 3)
		DrawEdge(hDC, &completeFillRect, EDGE_ETCHED, BF_RECT | BF_SOFT);
	else if (borderMethod == 4)
		DrawEdge(hDC, &completeFillRect, EDGE_BUMP, BF_RECT | BF_SOFT);
}

void TrayTexture::paintEffects(HDC destHDC, int width, int height)
{
	//Paint Saturation and Hueing Effects in MemoryDC's
	int iy = 0;
	int ix = 0;
	COLORREF cl = 0;
	BYTE r = 0;
	BYTE g = 0;
	BYTE b = 0;

	for(iy = 0; iy < height; iy++)
	{
		for(ix = 0; ix < width; ix++)
		{
			cl = GetPixel(destHDC, ix, iy);
			r = GetRValue(cl);
			g = GetGValue(cl);
			b = GetBValue(cl);
			
			if( !( r == 255 && g == 0 && b == 255 ) )
			{
				// Saturation effect
				if(taskbar_saturation != 255)
				{
					BYTE gray = (BYTE)(r * 0.3086 + g * 0.6094 + b * 0.0820);

					r = (BYTE)((r * taskbar_saturation + gray * (255 - taskbar_saturation) + 255) >> 8);
					g = (BYTE)((g * taskbar_saturation + gray * (255 - taskbar_saturation) + 255) >> 8);
					b = (BYTE)((b * taskbar_saturation + gray * (255 - taskbar_saturation) + 255) >> 8);
				}

				// Hue effect
				if(taskbar_hueIntensity != 0)
				{
					// The B & R values of the hue is swapped here, can somebody tell me why it only works that way?
					// test normal
					r = (BYTE)((r * (255 - taskbar_hueIntensity) + GetRValue(taskbar_hueColor) * taskbar_hueIntensity + 255) >> 8);
					g = (BYTE)((g * (255 - taskbar_hueIntensity) + GetGValue(taskbar_hueColor) * taskbar_hueIntensity + 255) >> 8);
					b = (BYTE)((b * (255 - taskbar_hueIntensity) + GetBValue(taskbar_hueColor) * taskbar_hueIntensity + 255) >> 8);
				}
			}
			else
			{
				transparencyNeeded = true;
			}

			SetPixelV(destHDC, ix, iy, RGB(r, g, b));
		}
	}
	//End Saturation / Hueing
}

void TrayTexture::solidGradientPaint(HDC destHDC, RECT r)
{
	if ( repeatedGradient == 0 )
	{
		int nWidth = r.right-r.left;
		int nHeight = r.bottom-r.top;

		RECT r2 = {0,0,0,0};
		double cr, cg, cb, dr, dg, db;

		// ColorsSteps between Start and End Color
		int gradSteps = 0;
		int unequalRest = 0;

		int colorCount = Gradient.size();

		if (verticalGradient)
		{
			gradSteps = (nHeight-colorCount)/(colorCount-1);
			unequalRest = (nHeight-colorCount)%(colorCount-1);
		}
		else
		{
			gradSteps = (nWidth-colorCount)/(colorCount-1);
			unequalRest = (nWidth-colorCount)%(colorCount-1);
		}

		int lastGradientPos = 0;
		if (verticalGradient)
			lastGradientPos = r.top;
		else
			lastGradientPos = r.left;

		int counter = 0;
		for (ColorRefListIterator cit = Gradient.begin(); cit != Gradient.end(); cit++ )
		{
			if (cit == Gradient.end())
				break;

			counter++;

			COLORREF one = *cit;
			cit++;
			COLORREF two = *cit;
			cit--;
		
			// get current color
			cr = GetRValue( one );
			cg = GetGValue( one );
			cb = GetBValue( one );

			// Get delta color (+1 cause we must reach the Final Color)
			dr = (cr - GetRValue( two )) / (gradSteps+1);
			dg = (cg - GetGValue( two )) / (gradSteps+1);
			db = (cb - GetBValue( two )) / (gradSteps+1);

			//Paint Real Gradient Color
			//-------------------------------------------
			if ( unequalRest != 0 && counter > 1 )
			{
				unequalRest--;
				if ( verticalGradient )
				{
					SetRect(&r2, r.left, lastGradientPos, r.right, min(lastGradientPos+1, r.bottom));
				}
				else
				{
					SetRect(&r2, lastGradientPos, r.top, min(lastGradientPos+1, r.right), r.bottom);
				}

				if ( r2.right-r2.left > 0 && r2.bottom-r2.top > 0 )
				{
					::SetBkColor(destHDC, one);
					::ExtTextOut(destHDC, 0, 0, ETO_OPAQUE, &r2, NULL, 0, NULL);
				}
				
				if ( verticalGradient )
					lastGradientPos = r2.bottom;
				else
					lastGradientPos = r2.right;
			}
			//-------------------------------------------

			for (int i = 0; i <= gradSteps; i++)
			{
				if ( verticalGradient )
				{
					SetRect(&r2, r.left, lastGradientPos+i, r.right, min(lastGradientPos+i+1, r.bottom));
				}
				else
				{
					SetRect(&r2, lastGradientPos+i, r.top, min(lastGradientPos+i+1, r.right), r.bottom);
				}

				if ( r2.right-r2.left > 0 && r2.bottom-r2.top > 0 )
				{
					::SetBkColor(destHDC, RGB((int)cr, (int)cg, (int)cb));
					::ExtTextOut(destHDC, 0, 0, ETO_OPAQUE, &r2, NULL, 0, NULL);
				}
				
				cr -= dr;
				cg -= dg;
				cb -= db;
			}

			if ( verticalGradient )
				lastGradientPos = r2.bottom;
			else
				lastGradientPos = r2.right;
		}
	}
	else
	{
		int nWidth = r.right-r.left;
		int nHeight = r.bottom-r.top;

		RECT r2 = {0,0,0,0};
		COLORREF color;

		int lastGradientPos = 0;
		if ( verticalGradient )
			lastGradientPos = r.top;
		else
			lastGradientPos = r.left;

		int counter = 0;
		int nSize = 0;
		
		if ( verticalGradient )
			nSize = nHeight;
		else
			nSize = nWidth; 

		while ( counter < nSize )
		{
			for (ColorRefListIterator cit = Gradient.begin(); cit != Gradient.end(); cit++ )
			{
				color = *cit;

				if ( verticalGradient )
				{
					SetRect(&r2, r.left, lastGradientPos, r.right, min(lastGradientPos+repeatedGradient, r.bottom));
				}
				else
				{
					SetRect(&r2, lastGradientPos, r.top, min(lastGradientPos+repeatedGradient, r.right), r.bottom);
				}

				if ( r2.right-r2.left > 0 && r2.bottom-r2.top > 0 )
				{
					::SetBkColor(destHDC, color);
					::ExtTextOut(destHDC, 0, 0, ETO_OPAQUE, &r2, NULL, 0, NULL);
				}

				if ( verticalGradient )
					lastGradientPos = r2.bottom;
				else
					lastGradientPos = r2.right;

				counter++;

				if ( counter == nSize )
					break;
			}
		}
	}
}

void TrayTexture::PrepareAlphaChannel(HBITMAP &bmp)
{
	//Paint Saturation and Hueing Effects in MemoryDC's
	BITMAP bm;
	GetObject(bmp, sizeof(BITMAP), &bm);

	BYTE *pPixel = (LPBYTE)bm.bmBits;

	BYTE r = 0;
	BYTE g = 0;
	BYTE b = 0;

	for (int i = 0; i < bm.bmHeight * bm.bmWidth; ++i)
	{
		//Magic Pink
		if (pPixel[0] == 255 && pPixel[1] == 0 && pPixel[2] == 255)
			pPixel[3] = 0;

		//Prepare Text
		if (pPixel[3] == 0)
			pPixel[3] = 1;

		r = pPixel[0];
		g = pPixel[1];
		b = pPixel[2];

		//AlphaMap Colorizing
		//Saturation Effect
		if( taskbar_saturation != 255 )
		{
			BYTE gray = (BYTE)(r * 0.3086 + g * 0.6094 + b * 0.0820);
			r = (BYTE)((r * taskbar_saturation + gray * (255 - taskbar_saturation) + 255) >> 8);
			g = (BYTE)((g * taskbar_saturation + gray * (255 - taskbar_saturation) + 255) >> 8);
			b = (BYTE)((b * taskbar_saturation + gray * (255 - taskbar_saturation) + 255) >> 8);
		}
		// Hue effect
		if( taskbar_hueIntensity != 0 )
		{
			// The B & R values of the hue is swapped here, can somebody tell me why it only works that way?
			// test normal
			b = (BYTE)((r * (255 - taskbar_hueIntensity) + GetRValue(taskbar_hueColor) * taskbar_hueIntensity + 255) >> 8);
			g = (BYTE)((g * (255 - taskbar_hueIntensity) + GetGValue(taskbar_hueColor) * taskbar_hueIntensity + 255) >> 8);
			r = (BYTE)((b * (255 - taskbar_hueIntensity) + GetBValue(taskbar_hueColor) * taskbar_hueIntensity + 255) >> 8);
		}
		
		//PreMultiply
		pPixel[0] = r * pPixel[3] / 255;
		pPixel[1] = g * pPixel[3] / 255;
		pPixel[2] = b * pPixel[3] / 255;

		pPixel+= 4;
	}
}
